// StudentGPA.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include "StudentGPA.h"
#include <time.h>
#include <fstream>
#include <afxtempl.h>
#include <cstdlib>
#include "GradeRecord.h"
#include "HighResolutionTime.h"

char* cacheFile = "cacheFile";

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// The one and only application object
CWinApp theApp;

bool isFileRefresh();
// Precondition:  None
// Postcondition: Returns true if cacheFile should be refreshed
//                Conditions for refresh are: 
//                   - cacheFile does not exist
//                   - cacheFile is more than 1 day old

void refreshFile();
// Precondition:  Relational database with student, course, and grade tables exits
// Postcondition: Creates a cacheFile representing a cached version of the relational
//                database.

void loadFile();
// Precondition:  cacheFile exists
// Postcondition: cacheFile is loaded into an in-memory data structure studentList

// List of all student grade records
CList <CGradeRecord, CGradeRecord&> studentList;

void displayMenu();
void gpaBySSN();
void gpaByLastName();
void gpaByCourseId();
void sort(CList <CGradeRecord, CGradeRecord&> & list);
void removeDuplicates(CList <CGradeRecord, CGradeRecord&> & list);
HighResolutionTime t;

using namespace std;

int _tmain(int argc, TCHAR* argv[], TCHAR* envp[])
{
	int nRetCode = 0;

	// initialize MFC and print and error on failure
	if (!AfxWinInit(::GetModuleHandle(NULL), NULL, ::GetCommandLine(), 0))
	{
		// Error!  This should never happen
		cerr << _T("Fatal Error: MFC initialization failed") << endl;
		nRetCode = 1;
	}

	//////////////////////////////////////////
	// This is where the real program starts
	//////////////////////////////////////////
	else
	{
		
		//////////////////////////////////////
		// INITIALIZATION
		//////////////////////////////////////
		// Refresh cache file if necessary
		if (isFileRefresh()) {
			refreshFile();
		}

		// Load the cacheFile into memory
		loadFile();

		//////////////////////////////////////
		// MENU
		/////////////////////////////////////
		displayMenu();

	}

	// Exit
	cout << "Press enter to terminate..." << endl;
	char c;
	cin.get(c);
	return nRetCode;
}

bool isFileRefresh()
{
	// Get time file was last modified
	CFileStatus status;
	int fileExists = CFile::GetStatus( cacheFile, status );

	if (!fileExists) return true;

	// Get current time
	CTime current = CTime::GetCurrentTime();

	// Compare day of current time to day of file
	return !(status.m_mtime.GetDay()   == current.GetDay() &&
			 status.m_mtime.GetMonth() == current.GetMonth() &&
			 status.m_mtime.GetYear()  == current.GetYear());
}

void refreshFile()
{
	cout << "Creating cache file." << endl;

	CDatabase db;
	
	BOOL result = db.Open(NULL,FALSE,FALSE,
						  _T("ODBC;DSN=CS325;UID=SCOTT;PWD=TIGER;"),
						  TRUE);

	if (!result) {
		cout << "Database Connection failed." << endl;
		exit(0);
	}

	// Use forwardOnly type recordset for best
	// performance, since only MoveNext is required
	CRecordset rs( &db );

	rs.Open( CRecordset::forwardOnly,
			 _T( "SELECT student.ssn, first, last, course.id, name, grade FROM student,course,grade where student.ssn=grade.ssn and course.id=grade.id" ) );

	// Loop through the recordset,
	// using GetFieldValue and
	// GetODBCFieldCount to retrieve
	// data in all columns
	short nFields = rs.GetODBCFieldCount( );

	cout << "Number of fields is: " << nFields << endl;

	cout << "Number of records is: " << rs.GetRecordCount() << endl;

	ofstream out;
	out.open("cacheFile");
	long count=0;
	while( !rs.IsEOF() ) {

		// Create a CDBVariant object to
		// store field data
		CDBVariant ssn;
		CDBVariant fname;
		CDBVariant lname;
		CDBVariant cid;
		CDBVariant cname;
		CDBVariant grade;

		rs.GetFieldValue( "SSN", ssn );
		rs.GetFieldValue( "FIRST", fname );
		rs.GetFieldValue( "LAST", lname );
		rs.GetFieldValue( "ID", cid );
		rs.GetFieldValue( "NAME", cname );
		rs.GetFieldValue( "GRADE", grade );
	

		out  << (LPCTSTR) (*ssn.m_pstring)
			 << "|" << (LPCTSTR) (*fname.m_pstring)
			 << "|" << (LPCTSTR) (*lname.m_pstring)
			 << "|" << (LPCTSTR) (*cid.m_pstring)
			 << "|" << (LPCTSTR) (*cname.m_pstring)
			 << "|" <<  (LPCTSTR) (*grade.m_pstring)
			 << endl; 

		rs.MoveNext();

		count++;
		if ((count % 1000) == 0) {
			cout << ".";
			cout.flush();
		}
	}

	cout << endl;

	rs.Close();
	db.Close();
	out.close();

	cout << "Cache file creation complete. " << endl;
}

void loadFile()
{
	cout << "Loading cache file into memory." << endl;

	ifstream in;
	in.open("cacheFile");

	int count=0;
	while (!in.fail()) {
		CGradeRecord* r = new CGradeRecord();
		in >> *r;	
		studentList.AddHead(*r);
		count++;
		if ((count % 1000) == 0) {
				cout << ".";
				cout.flush();
		}
	}
	cout << endl;

	int size=studentList.GetCount();

	cout << "Cache file loaded " << size << " records." << endl;
}

void displayMenu()
{
	while (true) {
		system("cls");
		cout << "Welcome to the Stonehill Student GPA Application" << endl;
		cout << endl;
		cout << "Select one of the following" << endl;
		cout << "----- Main Menu -----" << endl;
		cout << endl;
		cout << "1) GPA by SSN" << endl;
		cout << "2) GPA by Last Name" << endl;
		cout << "3) GPA by CourseID" << endl;
		cout << endl;
		cout << "0) Exit" << endl;
		cout << endl;
		cout << endl;
		cout << ">";

		char c;
		cin >> c;

		switch (c)
		{
			case '0':
				return;
				break;

			case '1':
				gpaBySSN();
				break;

			case '2':
				gpaByLastName();
				break;

			case '3':
				gpaByCourseId();
				break;

		}
	}
}

void gpaByLastName()
{
	string lname;
	char c;

	cout << "Enter student last name: " << endl;
	cin  >> lname;
	cin.get(c);
	
	int size=studentList.GetCount();
	POSITION head=studentList.GetHeadPosition();
	CList <CGradeRecord, CGradeRecord&> resultList;
	double gpa=0;
	int    count=0;

	for (int i=0; i<size; i++) {
		CGradeRecord r = studentList.GetNext(head);
		if (lname == r.lname) {
			resultList.AddHead(r);
			cout << ".";
			cout.flush();
			gpa += r.grade;
			count++;
		}
	}
	
	cout << endl;

	if (count != 0) {
		removeDuplicates(resultList);
		POSITION pos = resultList.GetHeadPosition();
		CGradeRecord r = resultList.GetHead();
		while (pos) {
			cout << r.ssn
				 << " "
				 << r.fname
				 << " "
				 << r.lname
				 << endl;
			r = resultList.GetNext(pos);
		}
		cout << "GPA is: " << gpa/count << endl;
	}
	else {
		cout << "Student not found!"  << endl;
	}

	cout << "Press enter to continue..." << endl;
	cin.get(c);
}

void gpaBySSN()
{
	string ssn;
	char c;

	cout << "Enter student SSN: " << endl;
	cin  >> ssn;
	cin.get(c);

	long start = t.getTimeMilliSeconds();

	int size=studentList.GetCount();
	POSITION head=studentList.GetHeadPosition();
	CList <CGradeRecord, CGradeRecord&> resultList;
	double gpa=0;
	int    count=0;

	for (int i=0; i<size; i++) {
		CGradeRecord r = studentList.GetNext(head);
		if (ssn == r.ssn) {
			resultList.AddHead(r);
			cout << ".";
			cout.flush();
			gpa += r.grade;
			count++;
		}
	}

	cout << endl;
	if (count != 0) {
		CGradeRecord r = resultList.GetHead();
		cout << r.ssn
			 << " "
			 << r.fname
			 << " "
			 << r.lname
			 << " "
			 << "GPA is: " 
			 << gpa/count 
			 << endl;
	}
	else {
		cout << "Student not found!"  << endl;
	}

	cout << "Press enter to continue..." << endl;
	cin.get(c);
}

void gpaByCourseId()
{
	string courseId;
	char c;

	cout << "Enter student course id (ccNNN): " << endl;
	cin  >> courseId;
	cin.get(c);

	int size=studentList.GetCount();
	POSITION head=studentList.GetHeadPosition();
	CList <CGradeRecord, CGradeRecord&> resultList;
	double gpa=0;
	int    count=0;

	for (int i=0; i<size; i++) {
		CGradeRecord r = studentList.GetNext(head);
		if (courseId == r.courseId) {
			resultList.AddHead(r);
			cout << ".";
			cout.flush();
			gpa += r.grade;
			count++;
		}
	}

	cout << endl;
	if (count != 0) {
		CGradeRecord r = resultList.GetHead();
		cout << r.courseId 
			 << " "
			 << r.courseName
			 << " "
			 << "GPA is: " 
			 << gpa/count 
			 << endl;
	}
	else {
		cout << "Course not found!"  << endl;
	}

	cout << "Press enter to continue..." << endl;
	cin.get(c);
}

void sort(CList <CGradeRecord, CGradeRecord&> & list) 
{
	// array location of last value exchanged
	// initialized to rightmost position in array

	int lastExchanged = list.GetCount()-1;
	
	while (lastExchanged > 0)
	{
		// All values to right of sorted_pos are are sorted in the array. 
		// As the algorithm executes this variable gets smaller and smaller 
		// approaching zero when it reaches the left side of the array.
		int sortedPos = lastExchanged;
		lastExchanged=0;
		
		// Starting at the beginning of the array.  Bubble up the largest
		// unsorted value in the array.  To do this, move the largest value from 
		// left to right by looking at array value pairs.  Swap pair values if
		// if the left is greater than right value.  Stop when you've reached the
		// sorted values in the array.

		for (int bubblePos=0; bubblePos<sortedPos; bubblePos++)
		{
			// swap pair values if left greater than right
			if ((list.GetAt(list.FindIndex(bubblePos))).ssn>(list.GetAt(list.FindIndex(bubblePos+1))).ssn)
			{
			// swap array[bubble_pos] and array[bubble_pos+1]
			CGradeRecord temp = list.GetAt(list.FindIndex(bubblePos));
			list.SetAt(list.FindIndex(bubblePos),list.GetAt(list.FindIndex(bubblePos+1)));
			list.SetAt(list.FindIndex(bubblePos+1),temp);
			//swap(array[bubblePos], array[bubblePos+1]);
		
			// When exiting the for loop all values from position lastExchanged
			// to size - 1 will be sorted.
			lastExchanged = bubblePos;
			}
		}
	}
}

void removeDuplicates(CList <CGradeRecord, CGradeRecord&> & list)
{
	sort(list);

	POSITION head = list.GetHeadPosition();
	CGradeRecord current = list.GetHead();

	if (!head) return;

	CGradeRecord next = list.GetNext(head);

	while (head) {
		if (current.ssn == next.ssn) {
			POSITION old = head;
			next = list.GetNext(head);
			list.RemoveAt(old);
		} else {
			current = next;
			next = list.GetNext(head);
		}

	}

}